/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.persistence.concurrency;

import org.objectweb.perseus.concurrency.pessimistic.PessimisticConcurrencyManager;
import org.objectweb.perseus.concurrency.pessimistic.Lock;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.persistence.api.StateManager;
import org.objectweb.perseus.persistence.api.StorageManager;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.api.State;
import org.objectweb.perseus.persistence.api.PersistenceException;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class PPessimisticConcurrencyManager
        extends PessimisticConcurrencyManager {

    public PPessimisticConcurrencyManager() throws ConcurrencyException {
    }

    public PPessimisticConcurrencyManager(short policy) throws ConcurrencyException {
        super(policy);
    }

    public final static String STORAGE_MANAGER_BINDING = "storage-manager";
    public final static String STATE_MANAGER_BINDING = "state-manager";

    protected StateManager stateManager;

    protected StorageManager storageManager;

    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
        return new String[] {
            STORAGE_MANAGER_BINDING,
            STATE_MANAGER_BINDING };
    }

    public Object lookupFc(String s) {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            return storageManager;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            return stateManager;
        } else {
            return super.lookupFc(s);
        }
    }

    public void bindFc(String s, Object o) {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = (StorageManager) o;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = (StateManager) o;
        } else super.bindFc(s, o);
    }

    public void unbindFc(String s) {
        if (STORAGE_MANAGER_BINDING.equals(s)) {
            storageManager = null;
        } else if (STATE_MANAGER_BINDING.equals(s)) {
            stateManager = null;
        } else {
            super.unbindFc(s);
        }
    }

	protected Object getState(Object ctx,
							  Object resourceId,
							  Lock lock,
							  Object hints) throws ConcurrencyException {
		CacheEntry ce = (CacheEntry) hints;
		WorkingSet ws = (WorkingSet) ctx;
        State state = stateManager.getReferenceState(ce);
        if (state == null) {
            synchronized (ce) {
                state = stateManager.getReferenceState(ce);
                if (state == null) {
                    state = stateManager.createState(ce);
                    stateManager.setReferenceState(state.getCacheEntry(), state);
                    try {
                        storageManager.read(ws, resourceId, state);
                        stateManager.makeClean(state);
                    } catch (PersistenceException e) {
                        stateManager.setReferenceState(state.getCacheEntry(), null);
                        stateManager.destroyState(state);
                        //The instance does not exist in the data support, remove the lock
                        synchronized (locks) {
                            if (lock.close(ctx)) {
                                locks.remove(lock);
                            }
                        }
                        throw new NoDSIConcurrencyException(e);
                    }
                }
            }
        }
        return state;
    }
}
